--[[
	Project: CombatAlerts
	Author: TimeBomb
	
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
	
]]

------------------------
--[[ PUBLIC MEMBERS ]] --
------------------------
CAlerts = {}

-------------------------
--[[ PRIVATE MEMBERS ]] --
-------------------------
local theAlert

-- @field table of unit detail information for players you have recently interacted with.
local unitDetailsCache = {}

-- @field holds who has damaged who, when and by how much
local recentDamage = {}

-- @field holds who has healed who, when and by how much
local recentHeals = {}

-- @field holds to-be-deleted profiles and text profiles
local profileDeleteQueue = {}
local textProfileDeleteQueue = {}

local playerDetail = nil

local defaultConfig = {
                        -- damageRatioSplit = The amount of damage an ally/enemy need to contribute towards the death of something for it to be counted as an assist
                        -- healingRatioSplit = The amount of healing an ally/enemy need to contribute towards the death of something for it to be counted as an assist
                        -- distractRatioSplit = The amount of damage an ally/enemy need to take while someone else dies for it to be counted as an assist
                        -- Note: All values are in percentage form, and account for the percentage of the dead ally/enemy's max HP (0.11 = 11%, 0.5 = 50%, 1 = 100%)
                        damageRatioSplit = 0.5,
                        healingRatioSplit = 0.5,
                        distractRatioSplit = 0.5,
                        timeAssistWindow = 10, -- The amount of time, in seconds, prior to the death, that assists will be accounted for.
                        toggle = 1, -- 0 = Off, 1 = On
                        mode = 0, -- 0 = All, 1 = PvE Only, 2 = PvP Only, 3 = Debug(Everything always enabled)
                        extendedalerts = false,
                        strings = {},
                        fonttitle = "title1.ttf",
                        fontmessage = "message1.ttf",
                        kbmode = 0,
                     }
       
local defaultStrings = {
                        title = {
                                    victory = {
                                                solokill = "Demolition",
                                                killingblow = "Final Blow",
                                                damageassist = "Tango Down",
                                                distractassist = "",
                                                healassist = "Soothing Hand",
                                              },
                                    death   = {
                                                solokill = "Fatality",
                                                killingblow = "Stricken Down",
                                                damageassist = "Ganked",
                                                distractassist = "Fatal Annoyance",
                                                healassist = "Blasphmey",
                                              },
                                },
                        message = {
                                    victory = {
                                                solokill = "You single-handedly demolished {1}",
                                                killingblow = "You finished off {1}",
                                                damageassist = "You helped slaughter {1}",
                                                distractassist = "",
                                                healassist = "You healed your allies while they defeated {1}",
                                              },
                                    death   = {
                                                solokill = "{1} beat you down",
                                                killingblow = "{1} finished you off",
                                                damageassist = "{1} contributed to your death",
                                                distractassist = "{1} distracted you for long enough",
                                                healassist = "{1} healed {2} whilst they killed you",
                                              },
                                  },
                       }

------------------------
--[[ PUBLIC METHODS ]] --
------------------------
if not CAlerts_Profiles then
    CAlerts_Profiles = {}
end

if not CAlerts_TextProfiles then
    CAlerts_TextProfiles = {}
end

if not CAlerts_Config then
    CAlerts_Config = defaultConfig
end

if not CAlerts_CurrentProfile then
    CAlerts_CurrentProfile = ""
end

if not CAlerts_CurrentTextProfile then
    CAlerts_CurrentTextProfile = ""
end

-------------------------
--[[ PRIVATE METHODS ]] --
-------------------------

-- Addon initialisation
-- 1. Registers event handlers to deal with damage, heal and death events.
-- 2. Removes all session information fromt the database.
-- 3. Outputs initialisation message.
local function init()

  -- create the alert
  theAlert = CAlerts.Alert()

  table.insert(Event.System.Update.Begin, { theAlert.Update, "CAlerts", "AlertUpdate" })

  -- @event register event to deal with assists and distracts
  table.insert(Event.Combat.Damage, { CAlerts.doDamage, "CAlerts", "CAlerts.doDamage" })

  -- @event register handler to deal with deaths
  table.insert(Event.Combat.Death, { CAlerts.doDeath, "CAlerts", "CAlerts.doDeath" })

  -- @event register handler to deal with heals
  table.insert(Event.Combat.Heal, { CAlerts.doHeal, "CAlerts", "CAlerts.doHeal" })

  -- @event register handler to deal with new units being seen
  table.insert(Event.Unit.Remove, { CAlerts.doNewUnit, "CAlerts", "CAlerts.doNewUnit" })

  -- Register the slash commands
  table.insert(Command.Slash.Register("calerts"), { CAlerts.commandHandler, "CAlerts", "Slash command" })
  
  if CAlerts_CurrentProfile ~= "" then
    if not CAlerts_Profiles[CAlerts_CurrentProfile] then
        print("Error - Current Profile: " .. CAlerts_CurrentProfile .. " not found. Falling back on default configs.")
        CAlerts_CurrentProfile = ""
    else
        CAlerts_Config["damageRatioSplit"] = CAlerts_Profiles[CAlerts_CurrentProfile]['dmgratio']
        CAlerts_Config["healingRatioSplit"] = CAlerts_Profiles[CAlerts_CurrentProfile]['healratio']
        CAlerts_Config["distractRatioSplit"] = CAlerts_Profiles[CAlerts_CurrentProfile]['distractratio']
        CAlerts_Config["timeAssistWindow"] = CAlerts_Profiles[CAlerts_CurrentProfile]['assist']
    end
  end
  
    if CAlerts_CurrentTextProfile ~= "" then
        if not CAlerts_TextProfiles[CAlerts_CurrentTextProfile] then
            print("Error - Current Text Profile: " .. CAlerts_CurrentTextProfile .. " not found. Falling back on default configs.")
            CAlerts_CurrentTextProfile = ""
        else
            CAlerts_Config["strings"] = CAlerts_TextProfiles[CAlerts_CurrentTextProfile]
        end
  end
  
  local next = next
  if next(CAlerts_Config["strings"]) == nil then
    CAlerts_Config["strings"] = defaultStrings
  end

  
  local welcomeTxt = "Combat Alerts loaded."
  if CAlerts_CurrentProfile ~= "" then
    welcomeTxt = welcomeTxt .. " Current Profile: " .. CAlerts_CurrentProfile .. "."
  end
  if CAlerts_CurrentTextProfile ~= "" then
    welcomeTxt = welcomeTxt .. " Current Text Profile: " .. CAlerts_CurrentTextProfile .. "."
  end
  print(welcomeTxt)
end

--
local isApiInitialised = false
local function isApiInitialisedLoop()
  if isApiInitialised then return end
  if Inspect.Unit.Detail('player') ~= nil then
    isApiInitialised = true
    init()
  end
end

-- the addon has loaded but the addon API is not fully initialised (annoying)
-- start a loop to check
local function addonLoaded()
  table.insert(Event.System.Update.Begin, { isApiInitialisedLoop, "CAlerts", "isApiInitialised" })
end

-- Called when there is a unit addevent
-- 1. Stores the unit detail information for the caster and target of the damage event for future use.
-- @param evt a table containing the following combat damage event information (as of Rift version 1.5): ability, abilityname, caster unit id, damage, target unit id.

function CAlerts.doNewUnit(evt)
  for i, _ in pairs(evt) do
    if i ~= false then
      -- store unit information into cache if required
      if unitDetailsCache[i] == nil then
        if Inspect.Unit.Detail(i) ~= nil then
          unitDetailsCache[i] = Inspect.Unit.Detail(i)
        end
      end
    end
  end
end


-- Called when there is a combat damage event
-- 1. Stores the event information into the recentDamage table for future analysis as and when a death event occurs.
-- 2. Stores the unit detail information for the caster and target of the damage event for future use.
-- @param evt a table containing the following combat damage event information (as of Rift version 1.5): ability, abilityname, caster unit id, damage, target unit id.
function CAlerts.doDamage(evt)

  -- get my unit id
  local myUnitId = Inspect.Unit.Lookup('player')

  -- get current timestamp
  local time = Inspect.Time.Real()

  -- store the event information
  recentDamage[time] = evt

  -- store unit information into cache if required
  if unitDetailsCache[evt.caster] == nil then
    if Inspect.Unit.Detail(evt.caster) ~= nil then
      unitDetailsCache[evt.caster] = Inspect.Unit.Detail(evt.caster)
    end
  end
  if unitDetailsCache[evt.target] == nil then
    if Inspect.Unit.Detail(evt.target) ~= nil then
      unitDetailsCache[evt.target] = Inspect.Unit.Detail(evt.target)
    end
  end

  -- clean out recentDamage table of old records
  recentDamage = CAlerts.cleanOld(recentDamage)
end

-- Called when there is a combat heal event
-- 1. Stores the event information into the recentHeal table for future analysis as and when a death event occurs.
-- 2. Stores the unit detail information for the caster and target of the dhealge event for future use.
-- @param evt a table containing the following combat heal event information (as of Rift version 1.5): ability, abilityname, caster unit id, heal, target unit id.
function CAlerts.doHeal(evt)

  -- get my unit id
  local myUnitId = Inspect.Unit.Lookup('player')

  -- get current timestamp
  local time = Inspect.Time.Real()

  -- store the event information
  recentHeals[time] = evt

  -- store unit information into cache if required
  if unitDetailsCache[evt.caster] == nil then
    if Inspect.Unit.Detail(evt.caster) then
      unitDetailsCache[evt.caster] = Inspect.Unit.Detail(evt.caster)
    end
  end
  if unitDetailsCache[evt.target] == nil then
    if Inspect.Unit.Detail(evt.target) then
      unitDetailsCache[evt.target] = Inspect.Unit.Detail(evt.target)
    end
  end

  -- clean out recentDamage table of old records
  recentHeals = CAlerts.cleanOld(recentHeals)
end


-- Called when a death occurs
-- 1. Get the unit information for the caster and target (from the cache if needs be).
-- 2. Check we don't have missing information and that we did not kill ourselves.
-- 3. Calculate the damage,distract,heal percentage of the player against the target.
-- 4. Store data into database.
-- @param evt a table containing the death event information (as of Rift 1.5): caster unit id, target unit id.
function CAlerts.doDeath(evt)

  -- get killer and deceased
  local killerId = evt.caster
  local deceasedId = evt.target

  -- get the player id
  local playerId = Inspect.Unit.Lookup('player')

  -- hack to deal with duels and missing killers.
  if not killerId and deceasedId==playerId then
    --print("No killer Id, assuming player.target")
    killerId = Inspect.Unit.Lookup('player.target')
  end
  
  -- If you kill yourself(ex. fall to your death), it won't show up in the alerts
  if killerId == deceasedId then
    return
  end

  -- get details and fallback to cache
  local killerDetail = false
  local deceasedDetail = false
  if killerId ~= nil then killerDetail = Inspect.Unit.Detail(killerId) end
  if deceasedId ~= nil then deceasedDetail = Inspect.Unit.Detail(deceasedId) end
  
  if not killerDetail then killerDetail = unitDetailsCache[killerId] end
  if not deceasedDetail then deceasedDetail = unitDetailsCache[deceasedId] end

  -- we can't record this event if we are missing information
  if not killerDetail then killerDetail = CAlerts.getUnknown() end
  if not deceasedDetail then deceasedDetail = CAlerts.getUnknown() end

  -- Victory! Work out player contribution against deceased.
  if deceasedId ~= playerId then

    -- allies got a kill
    
    if deceasedDetail.player and CAlerts_Config["mode"] == 1 then -- Don't show player kills if PvE mode is toggled
        return
    elseif not deceasedDetail.player and CAlerts_Config["mode"] == 2 then -- Don't show non-player kills if PvP mode is toggled
        return
    end

    local damageByPlayerToDeceased = 0
    local damageByEveryoneToDeceased = 0
    local damageByDeceasedToPlayer = 0
    local damageByDeceasedToEverybody = 0

    local healsByPlayerToKiller = 0
    local healsByEveryoneToKiller = 0

    for i, v in pairs(recentDamage) do

      -- get damager and damaged
      local damager = v.caster
      local damaged = v.target

      -- calculate total raw damage
      local damage = 0
      if v.damage then damage = damage + v.damage end
      if v.damageAbsorbed then damage = damage + v.damageAbsorbed end
      if v.damageBlocked then damage = damage + v.damageBlocked end
      if v.damageDeflected then damage = damage + v.damageDeflected end
      if v.damageModified then damage = damage + v.damageModified end

      -- player does the damage to deceased
      if damager == playerId and damaged == deceasedId then
        damageByPlayerToDeceased = damageByPlayerToDeceased + damage
      elseif
        -- deceased does damage to the player
      damager == deceasedId and damaged == playerId then
        damageByDeceasedToPlayer = damageByDeceasedToPlayer + damage
      end

      -- if anyone does damage to the deceased
      if damaged == deceasedId then
        damageByEveryoneToDeceased = damageByEveryoneToDeceased + damage
      elseif
        -- if the deceased does damage to anyone
      damager == deceasedId then
        damageByDeceasedToEverybody = damageByDeceasedToEverybody + damage
      end
    end


    for i, v in pairs(recentHeals) do

      -- get healer and healed
      local healer = v.caster
      local healed = v.target

      -- calculate total raw healing
      local heal = 0
      if v.heal then heal = heal + v.heal end

      -- player does the heal to killer
      if healer == playerId and healed == killerId then
        healsByPlayerToKiller = healsByPlayerToKiller + heal
      end

      -- anyone does healing to the killer
      if healed == killerId then
        healsByEveryoneToKiller = healsByEveryoneToKiller + heal
      end
    end

    -- calculate contributions

    -- player damage ratio: player MUST do damage to the target
    local playerDamageRatio = 0
    if damageByPlayerToDeceased > 0 then
      playerDamageRatio = damageByPlayerToDeceased / damageByEveryoneToDeceased
    else
      playerDamageRatio = 0
    end

    -- target must do damage to the player
    local playerDistractRatio = 0
    if damageByDeceasedToPlayer > 0 then
      playerDistractRatio = damageByDeceasedToPlayer / damageByDeceasedToEverybody
    else
      playerDistractRatio = 0
    end

    -- player must heal killer
    local playerHealRatio = 0
    if healsByPlayerToKiller > 0 then
      playerHealRatio = healsByPlayerToKiller / healsByEveryoneToKiller
    else
      playerHealRatio = 0
    end
    local vKillingBlow = (killerId == playerId) and 1 or 0
    local vSolo = (vKillingBlow == 1 and (playerDamageRatio == 1 or damageByEveryoneToDeceased == 0) and (playerDistractRatio == 1 or damageByDeceasedToEverybody == 0) and (playerHealRatio == 1 or healsByEveryoneToKiller == 0)) and 1 or 0
    local vDamageAssist = playerDamageRatio > CAlerts_Config["damageRatioSplit"] and 1 or 0
    local vDistractAssist = playerDistractRatio > CAlerts_Config["distractRatioSplit"] and 1 or 0
    local vHealAssist = playerHealRatio > CAlerts_Config["healingRatioSplit"] and 1 or 0

    -- check for solo or assist/killingblow
    local iconFlags = {}
    if vSolo == 1 and not ptSolo then
      table.insert(iconFlags, "SoloKill")
    end
    if vKillingBlow == 1 then
      table.insert(iconFlags, "KillingBlow")
    end
    if vDamageAssist == 1 then
      table.insert(iconFlags, "DamageAssist")
    end
    if vDistractAssist == 1 then
      table.insert(iconFlags, "DistractAssist")
    end
    if vHealAssist == 1 then
      table.insert(iconFlags, "HealAssist")
    end

    if vSolo == 1 then
      CAlerts.processAlert("solokill", true, {deceasedDetail.name}, iconFlags)
    else
      if not CAlerts_Config["extendedalerts"] then
          if vKillingBlow == 1 then
            CAlerts.processAlert("killingblow", true, {deceasedDetail.name}, iconFlags)
          elseif vHealAssist == 1 and vDamageAssist == 0 then
            CAlerts.processAlert("healassist", true, {deceasedDetail.name}, iconFlags)
          elseif vDamageAssist == 1 or vDistractAssist == 1 then
            CAlerts.processAlert("damageassist", true, {deceasedDetail.name}, iconFlags)
          end
      else
          local alertShown = false
          if vKillingBlow == 1 then
            if alertShown then
                CAlerts.processAlert("killingblow", true, {deceasedDetail.name}, iconFlags, true)
            else
                CAlerts.processAlert("killingblow", true, {deceasedDetail.name}, iconFlags)
                alertShown = true
            end
          end
          if vHealAssist == 1 and vDamageAssist == 0 then
            if alertShown then
                CAlerts.processAlert("healassist", true, {deceasedDetail.name}, iconFlags, true)
            else
                CAlerts.processAlert("healassist", true, {deceasedDetail.name}, iconFlags)
                alertShown = true
            end
          end
          if vDamageAssist == 1 or vDistractAssist == 1 then
            if alertShown then
                CAlerts.processAlert("damageassist", true, {deceasedDetail.name}, iconFlags, true)
            else
                CAlerts.processAlert("damageassist", true, {deceasedDetail.name}, iconFlags)
                alertShown = true
            end
          end
      end
    end
  else
    -- Enemies got a kill
    -- Unlike CAlerts, we will show all deaths, as I believe that knowing what killed you is always important. (And nothing is being logged.)
    
    local damageByEnemyToDeceasedPlayer = {}
    local damagebyAllEnemiesToDeceasedPlayer = 0
    local damageByDeceasedPlayerToEnemy = {}
    local damageByDeceasedPlayerToAllEnemies = 0

    local healingOfKillerByEnemy = {}
    local healingOfKillerByAllEnemies = 0

    for i, v in pairs(recentDamage) do

      -- get damager and damaged
      local damager = v.caster
      local damaged = v.target

      -- calculate total raw damage
      local damage = 0
      if v.damage then damage = damage + v.damage end
      if v.damageAbsorbed then damage = damage + v.damageAbsorbed end
      if v.damageBlocked then damage = damage + v.damageBlocked end
      if v.damageDeflected then damage = damage + v.damageDeflected end
      if v.damageModified then damage = damage + v.damageModified end

      if not damageByEnemyToDeceasedPlayer[damager] or damageByEnemyToDeceasedPlayer[damager] == nil then damageByEnemyToDeceasedPlayer[damager] = 0 end
      if not damageByDeceasedPlayerToEnemy[damaged] or damageByDeceasedPlayerToEnemy[damaged] == nil then damageByDeceasedPlayerToEnemy[damaged] = 0 end

      if damaged == playerId then
        -- damage done to player
        damageByEnemyToDeceasedPlayer[damager] = damageByEnemyToDeceasedPlayer[damager] + damage
        damagebyAllEnemiesToDeceasedPlayer = damagebyAllEnemiesToDeceasedPlayer + damage
      elseif damager == playerId then
        -- deceased does damage to the player
        damageByDeceasedPlayerToEnemy[damaged] = damageByDeceasedPlayerToEnemy[damaged] + damage
        damageByDeceasedPlayerToAllEnemies = damageByDeceasedPlayerToAllEnemies + damage
      end
    end

    if deceasedId == playerId then
      for i, v in pairs(recentHeals) do

        -- get healer and healed
        local healer = v.caster
        local healed = v.target

        if healer and healed then

          -- calculate total raw healing
          local heal = 0
          if v.heal then heal = heal + v.heal end

          -- anyone does healing to the killer
          if healed == killerId and deceasedId == playerId then
            if not healingOfKillerByEnemy[healer] then
              healingOfKillerByEnemy[healer] = 0
            end
            healingOfKillerByEnemy[healer] = healingOfKillerByEnemy[healer] + heal
            healingOfKillerByAllEnemies = healingOfKillerByAllEnemies + heal
          end
        end
      end
    end

    -- boolean variables to ensure we don't assign the same reward twice
    local pSolo = false
    local pKillingBlow = false
    local pDamageAssist = false
    local pDistractAssist = false
    local pHealAssist = false

    -- loop through all enemies that have healed the killer of the player
    local healerUnitId = false
    for i, _ in pairs(healingOfKillerByEnemy) do

      local enemyHealRatio = 0
      if healingOfKillerByEnemy[i] ~= nil then
        enemyHealRatio = healingOfKillerByEnemy[i] / healingOfKillerByAllEnemies
      end
      local dHealAssist = (enemyHealRatio > CAlerts_Config["healingRatioSplit"] and dSolo ~= 1) and 1 or 0
      local healerDetail = Inspect.Unit.Detail(i)
      if not healerDetail then healerDetail = unitDetailsCache[i] end
      if not healerDetail then
        healerDetail = CAlerts.getUnknown()
      end
      if dHealAssist == 1 and not pHealAssist then
        pHealAssist = true
        healerUnitId = i
        CAlerts.processAlert("healassist", false, {healerDetail.name, killerDetail.name}, {}, true)
        --nameArray = {healerDetail.name, killerDetail.name}
        --print("{" .. CAlerts.toAlertTitle(9) .. "} " .. CAlerts.toAlertMessage(nameArray, 9))
      end
    end -- end of for loop

    -- loop through all enemies that have damaged player and work out their contribution
    for i, _ in pairs(damageByEnemyToDeceasedPlayer) do

      local enemyDamageRatio = 0
      if damageByEnemyToDeceasedPlayer[i] ~= nil then
        enemyDamageRatio = damageByEnemyToDeceasedPlayer[i] / damagebyAllEnemiesToDeceasedPlayer
      end
      local enemyDistractRatio = 0
      if damageByDeceasedPlayerToEnemy[i] ~= nil then
        enemyDistractRatio = damageByDeceasedPlayerToEnemy[i] / damageByDeceasedPlayerToAllEnemies
      end

      local dSolo = (enemyDamageRatio == 1 and enemyDistractRatio == 1 and healerUnitId == i) and 1 or 0
      local dDamageAssist = (enemyDamageRatio > CAlerts_Config["damageRatioSplit"] and dSolo ~= 1) and 1 or 0
      local dDistractAssist = (enemyDistractRatio > CAlerts_Config["distractRatioSplit"] and dSolo ~= 1) and 1 or 0
      local dKillingBlow = (killerId == i) and 1 or 0

      local damagerDetail = Inspect.Unit.Detail(i)
      if not damagerDetail then damagerDetail = unitDetailsCache[i] end
      if not damagerDetail then
        damagerDetail = CAlerts.getUnknown()
      end

      -- check for solo or assist/killingblow
      if dSolo == 1 and not pSolo and killerId ~= deceasedId then
        pSolo = true
        CAlerts.processAlert("solokill", false, {killerDetail.name}, {'SoloKill'})
      else
        if not CAlerts_Config["extendedalerts"] then
            if dKillingBlow == 1 and not pKillingBlow  and killerId ~= deceasedId then
              pKillingBlow = true
              CAlerts.processAlert("killingblow", false, {killerDetail.name}, {'KillingBlow'})
            elseif dDamageAssist == 1 and not pDamageAssist  and killerId ~= deceasedId then
              pDamageAssist = true
              CAlerts.processAlert("damageassist", false, {damagerDetail.name})
            elseif dDistractAssist == 1 and not pDistractAssist  and killerId ~= deceasedId then
              pDistractAssist = true
              CAlerts.processAlert("distractassist", false, {damagerDetail.name})
            end
        else
            local alertShown = false
            if dKillingBlow == 1 and not pKillingBlow  and killerId ~= deceasedId then
              pKillingBlow = true
              if alertShown then
                CAlerts.processAlert("killingblow", false, {killerDetail.name}, {'KillingBlow'}, true)
              else
                CAlerts.processAlert("killingblow", false, {killerDetail.name}, {'KillingBlow'})
                alertShown = true
              end
            end
            if dDamageAssist == 1 and not pDamageAssist  and killerId ~= deceasedId then
              pDamageAssist = true
              if alertShown then
                CAlerts.processAlert("damageassist", false, {damagerDetail.name}, {}, true)
              else
                CAlerts.processAlert("damageassist", false, {damagerDetail.name}, {})
                alertShown = true
              end
            end
            if dDistractAssist == 1 and not pDistractAssist  and killerId ~= deceasedId then
              pDistractAssist = true
              if alertShown then
                CAlerts.processAlert("distractassist", false, {damagerDetail.name}, {}, true)
              else
                CAlerts.processAlert("distractassist", false, {damagerDetail.name}, {})
                alertShown = true
              end
            end
        end
      end
    end -- end of for loop
  end
end


-- This function returns a template for Unit information when we don't have any information about the unit.
-- return table consisting of unknown unit information.
function CAlerts.getUnknown()
  local un = {}
  un.name = "Unknown Enemy"
  un.calling = "unknown"
  un.role = "unknown"
  un.level = "unknown"
  un.player = false
  return un
end

-- Remove anything older than assistWindow seconds
-- @param recents a table of recent damage, assiset, distract or heal events
-- @return table of events that are within the time window, anything older has been removed.
function CAlerts.cleanOld(recents)
  local currentTime = Inspect.Time.Real()
  local intime = {}
  for i, v in pairs(recents) do
    if i + CAlerts_Config["timeAssistWindow"] > currentTime then
      intime[i] = v
    end
  end
  return intime
end

-- Checks if we have a recent entry for the target
-- @param string targetName name of the target
-- @param table recents a table of recent interactions
-- @return the target entry or false if none is found
function CAlerts.checkRecent(targetUnitId, recents)
  local cleaned = CAlerts.cleanOld(recents)
  for i, v in pairs(cleaned) do
    if targetUnitId == v then
      return v
    end
  end
  return false
end


function CAlerts.commandHandler(args)
    local cmd = {}
    local numargs = 0
    for token in string.gmatch(args, "[^%s]+") do
        cmd[numargs] = token
        numargs = numargs + 1
    end
    if numargs > 0 then
        cmd[0] = cmd[0]:lower()
    end
    
    local helpString = "Possible commands:\n/calerts {on/off}\n/calerts mode {all/pvp/pve}\n/calerts {loadprofile/saveprofile/deleteprofile} value\n/calerts modify {damageratio/healratio/distractratio/assist} value\n/calerts defaults\n/calerts {loadtextprofile/savetextprofile/deletetextprofile} value\n/calerts modifytext {title/message} {victory/death} {solokill/killingblow/damageassist/distractassist/healassist} value\n/calerts defaulttext\n/calerts extendedalerts {on/off}\n/calerts loadpreset {pvpconfig/pveconfig/altconfig/calertstext}\n/calerts setfont {title/message} {1/2/3}\n/calerts kbonly {on/off}\n/calerts help"
    if numargs == 0 then
        print("Type '/calerts help' for a list of possible commands. For more info, look at DOCUMENTATION.TXT in the Combat Alerts addon folder.")
        return
    end
    
    if cmd[0] == "on" then
        if CAlerts_Config["toggle"] == 0 then -- Turn on
          CAlerts_Config["toggle"] = 1
          print("Combat Alerts has been turned on.")
        else
          print("Combat Alerts is already on.")
        end
    elseif cmd[0] == "off" then
        if CAlerts_Config["toggle"] == 0 then
          print("Combat Alerts is already off.")
        else -- Turn off
          CAlerts_Config["toggle"] = 0
          print("Combat Alerts has been turned off.")
        end
    elseif cmd[0] == "mode" then
        if cmd[1] == "all" then
            CAlerts_Config["mode"] = 0
            print("Combat Alerts has been set to show both PvE and PvP encounters.")
        elseif cmd[1] == "pve" then
            CAlerts_Config["mode"] = 1
            print("Combat Alerts has been set to show both only PvE encounters.")
        elseif cmd[1] == "pvp" then
            CAlerts_Config["mode"] = 2
            print("Combat Alerts has been set to show both only PvP encounters.")
        else
            print("Wrong Syntax. '/calerts mode {all/pvp/pve}'")
        end
    elseif cmd[0] == "saveprofile" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts saveprofile ProfileNameHere'")
        else
            local profileName = cmd[1]
            CAlerts_Profiles[profileName] = {}
            CAlerts_Profiles[profileName]['dmgratio'] = CAlerts_Config["damageRatioSplit"]
            CAlerts_Profiles[profileName]['healratio'] = CAlerts_Config["healingRatioSplit"]
            CAlerts_Profiles[profileName]['distractratio'] = CAlerts_Config["distractRatioSplit"]
            CAlerts_Profiles[profileName]['assist'] = CAlerts_Config["timeAssistWindow"]
            CAlerts_CurrentProfile = profileName
            print("Profile saved as " .. profileName)
        end
    elseif cmd[0] == "loadprofile" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts loadprofile ProfileNameHere'")
        else
            local profileName = cmd[1]
            if not CAlerts_Profiles[profileName] then
                print("Profile not found.")
            else
                CAlerts_Config["damageRatioSplit"] = CAlerts_Profiles[profileName]['dmgratio']
                CAlerts_Config["healingRatioSplit"] = CAlerts_Profiles[profileName]['healratio']
                CAlerts_Config["distractRatioSplit"] = CAlerts_Profiles[profileName]['distractratio']
                CAlerts_Config["timeAssistWindow"] = CAlerts_Profiles[profileName]['assist']
                CAlerts_CurrentProfile = profileName
                print("Profile: " .. profileName .. " loaded successfully.")
            end
        end
    elseif cmd[0] == "deleteprofile" then
        local thisHelpMsg = "Wrong Syntax. '/calerts deleteprofile ProfileNameHere'"
        if numargs ~= 2 and numargs ~= 3 then
            print(thisHelpMsg)
        else
            local profileName = cmd[1]
            if not CAlerts_Profiles[profileName] then
                print("Profile: " .. profileName .. " not found.")
            elseif numargs ~= 3 then
                profileDeleteQueue[profileName] = true
                print("You are about to delete profile: " .. profileName .. ". Please use the command '/calerts deleteprofile " .. profileName .. " confirm' to confirm profile deletion. Use '/calerts deleteprofile " .. profileName .. " cancel' to cancel deletion.")
            elseif cmd[2] == "confirm" then
                if profileDeleteQueue[profileName] ~= true then
                    print("This profile is not ready to be deleted. Please use the following syntax to delete this profile: '/calerts deleteprofile ProfileNameHere'")
                else
                    CAlerts_Profiles[profileName] = nil
                    profileDeleteQueue[profileName] = nil
                    print("Profile: " .. profileName .. " has been deleted successfully.")
                end
            elseif cmd[2] == "cancel" then
                if profileDeleteQueue[profileName] ~= true then
                    print("You have attempted to delete profile:" .. profileName .. ".")
                else
                    profileDeleteQueue[profileName] = nil
                    print("Profile: " .. profileName .. "'s deletion has been cancelled.")
                end
            else
                print(thisHelpMsg)
            end
        end
    elseif cmd[0] == "modify" then
        local thisHelpMsg = "Wrong Syntax. '/calerts modify {damageratio/healratio/distractratio/assist} value'"
        if numargs ~= 3 then
            print(thisHelpMsg)
        else
            if not tonumber(cmd[2]) then
                print("Error! Value must be integer.")
            elseif cmd[1] == "damageratio" then
                CAlerts_Config["damageRatioSplit"] = tonumber(cmd[2])
                CAlerts_CurrentProfile = ""
                print("Damage Ratio Split successfully set to: " .. cmd[2])
            elseif cmd[1] == "healratio" then
                CAlerts_Config["healingRatioSplit"] = tonumber(cmd[2])
                CAlerts_CurrentProfile = ""
                print("Healing Ratio Split successfully set to: " .. cmd[2])
            elseif cmd[1] == "distractratio" then
                CAlerts_Config["distractRatioSplit"] = tonumber(cmd[2])
                CAlerts_CurrentProfile = ""
                print("Distract Ratio Split successfully set to: " .. cmd[2])
            elseif cmd[1] == "assist" then
                CAlerts_Config["timeAssistWindow"] = tonumber(cmd[2])
                CAlerts_CurrentProfile = ""
                print("Assist Time Window successfully set to: " .. cmd[2])
            else
                print(thisHelpMsg)
            end
        end
    elseif cmd[0] == "defaults" then
        CAlerts_Config["damageRatioSplit"] = defaultConfig["damageRatioSplit"]
        CAlerts_Config["healRatioSplit"] = defaultConfig["healRatioSplit"]
        CAlerts_Config["distractRatioSplit"] = defaultConfig["distractRatioSplit"]
        CAlerts_Config["timeAssistWindow"] = defaultConfig["timeAssistWindow"]
        CAlerts_CurrentProfile = ""
        print("Reset configs back to default.")
    elseif cmd[0] == "debugalert" or cmd[0] == "debugprint" then
        if numargs < 4 then
            print("Error. Requires 4 arguments.")
        else
            if cmd[2] == "false" then cmd[2] = false end
            if not cmd[4] then
                if cmd[0] == "debugalert" then
                    CAlerts.processAlert(cmd[1], cmd[2], {cmd[3]})
                else
                    CAlerts.processAlert(cmd[1], cmd[2], {cmd[3]}, {}, true)
                end
            else
                if cmd[0] == "debugalert" then
                    CAlerts.processAlert(cmd[1], cmd[2], {cmd[3], cmd[4]})
                else
                    CAlerts.processAlert(cmd[1], cmd[2], {cmd[3], cmd[4]}, true)
                end
            end
        end
    elseif cmd[0] == "savetextprofile" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts savetextprofile ProfileNameHere'")
        else
            local textProfileName = cmd[1]
            CAlerts_TextProfiles[textProfileName] = CAlerts_Config["strings"]
            CAlerts_CurrentTextProfile = textProfileName
            print("Profile saved as " .. textProfileName)
        end
    elseif cmd[0] == "loadtextprofile" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts loadtextprofile ProfileNameHere'")
        else
            local textProfileName = cmd[1]
            if not CAlerts_TextProfiles[textProfileName] then
                print("Profile not found.")
            else
                CAlerts_Config["strings"] = CAlerts_TextProfiles[textProfileName]
                CAlerts_CurrentTextProfile = textProfileName
                print("Text Profile: " .. textProfileName .. " loaded successfully.")
            end
        end
    elseif cmd[0] == "deletetextprofile" then
        local thisHelpMsg = "Wrong Syntax. '/calerts deletetextprofile ProfileNameHere'"
        if numargs ~= 2 and numargs ~= 3 then
            print(thisHelpMsg)
        else
            local textProfileName = cmd[1]
            if not CAlerts_TextProfiles[textProfileName] then
                print("Profile: " .. textProfileName .. " not found.")
            elseif numargs ~= 3 then
                textProfileDeleteQueue[textProfileName] = true
                print("You are about to delete text profile: " .. textProfileName .. ". Please use the command '/calerts deletetextprofile " .. textProfileName .. " confirm' to confirm text profile deletion. Use '/calerts deletetextprofile " .. textProfileName .. " cancel' to cancel deletion.")
            elseif cmd[2] == "confirm" then
                if textProfileDeleteQueue[textProfileName] ~= true then
                    print("This text profile is not ready to be deleted. Please use the following syntax to delete this text profile: '/calerts deletetextprofile ProfileNameHere'")
                else
                    CAlerts_TextProfiles[textProfileName] = nil
                    textProfileDeleteQueue[textProfileName] = nil
                    print("Profile: " .. textProfileName .. " has been deleted successfully.")
                end
            elseif cmd[2] == "cancel" then
                if textProfileDeleteQueue[textProfileName] ~= true then
                    print("You have attempted to delete text profile:" .. textProfileName .. ".")
                else
                    textProfileDeleteQueue[textProfileName] = nil
                    print("Profile: " .. textProfileName .. "'s deletion has been cancelled.")
                end
            else
                print(thisHelpMsg)
            end
        end
    elseif cmd[0] == "modifytext" then
        if (numargs < 4) or (cmd[1] ~= 'title' and cmd[1] ~= 'message') or (cmd[2] ~= 'victory' and cmd[2] ~= 'death') or (cmd[3] ~= 'solokill' and cmd[3] ~= 'killingblow' and cmd[3] ~= 'damageassist' and cmd[3] ~= 'distractassist' and cmd[3] ~= 'healassist') then
            print("Wrong Syntax. '/calerts modifytext {title/message} {victory/death} {solokill/killingblow/damageassist/distractassist/healassist} value'")
        else
            local paramA = cmd[1]
            local paramB = cmd[2]
            local paramC = cmd[3]
            for i = 1, 3, 1 do table.remove(cmd, 1) end
            message = table.concat(cmd, " ")
            CAlerts_Config["strings"][paramA][paramB][paramC] = message
            print("Selected text successfully changed to: \"" .. message .. "\"")
        end
    elseif cmd[0] == "defaulttext" then
        CAlerts_Config["strings"] = defaultStrings
        CAlerts_CurrentTextProfile = ""
        print("Reset alert text back to default.")
    elseif cmd[0] == "loadpreset" then
        if numargs ~= 2 or (cmd[1] ~= "pvpconfig" and cmd[1] ~= "pveconfig" and cmd[1] ~= "altconfig" and cmd[1] ~= "calertstext") then
            print("Wrong Syntax. '/calerts loadpreset {pvpconfig/pveconfig/altconfig/calertstext}'")
        else
            if cmd[1] == "pvpconfig" then
                CAlerts_Config["damageRatioSplit"] = 0.39
                CAlerts_Config["healingRatioSplit"] = 0.4
                CAlerts_Config["distractRatioSplit"] = 0.5
                CAlerts_Config["timeAssistWindow"] = 8
                CAlerts_Config["mode"] = 2
                CAlerts_CurrentProfile = ""
                print("Now using PvP configuration preset.")
            elseif cmd[1] == "pveconfig" then
                CAlerts_Config["damageRatioSplit"] = 0.35
                CAlerts_Config["healingRatioSplit"] = 0.5
                CAlerts_Config["distractRatioSplit"] = 0.51
                CAlerts_Config["timeAssistWindow"] = 15
                CAlerts_Config["mode"] = 1
                CAlerts_CurrentProfile = ""
                print("Now using PvE configuration preset.")
            elseif cmd[1] == "altconfig" then
                CAlerts_Config["damageRatioSplit"] = 0.3
                CAlerts_Config["healingRatioSplit"] = 0.33
                CAlerts_Config["distractRatioSplit"] = 0.4
                CAlerts_Config["timeAssistWindow"] = 12
                CAlerts_Config["mode"] = 0
                CAlerts_CurrentProfile = ""
                print("Now using alternative configuration preset.")
            elseif cmd[1] == "botftext" then
                local botfPresetText =
                                        {
                                            title =   {
                                                        victory = {
                                                                    solokill = "Solo Kill",
                                                                    killingblow = "Killing Blow",
                                                                    damageassist = "Assisted Kill",
                                                                    distractassist = "",
                                                                    healassist = "Assisted Kill",
                                                                  },
                                                        death   = {
                                                                    solokill = "Solo Kill",
                                                                    killingblow = "Killing Blow",
                                                                    damageassist = "Damage Assist",
                                                                    distractassist = "Distract Assist",
                                                                    healassist = "Heal Assist",
                                                                  },
                                                      },
                                            message = {
                                                        victory = {
                                                                    solokill = "You single-handedly crushed {1}",
                                                                    killingblow = "You got the killing blow on {1}",
                                                                    damageassist = "You assisted in the slaughter of {1}",
                                                                    distractassist = "",
                                                                    healassist = "You assisted in the slaughter of {1}",
                                                                  },
                                                        death   = {
                                                                    solokill = "{1} single-handedly crushed you",
                                                                    killingblow = "{1} got the killing blow on you",
                                                                    damageassist = "{1} assisted in your slaughter",
                                                                    distractassist = "{1} distracted you whilst you were slaughtered",
                                                                    healassist = "{1} healed {2} whilst they killed you",
                                                                  },
                                                    },
                                        }
                CAlerts_CurrentTextProfile = ""
                CAlerts_Config["strings"] = botfPresetText
                print("Now using Book of the Fallen text presets.")
            end
        end
    elseif cmd[0] == "extendedalerts" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts extendedalerts {on/off}'")
        else
            if cmd[1] == "off" then
                if not CAlerts_Config["extendedalerts"] then
                    print("Extended Alerts is already off.")
                else
                    CAlerts_Config["extendedalerts"] = false
                    print("Extended Alerts has been turned off.")
                end
            else
                if CAlerts_Config["extendedalerts"] == true then
                    print("Extended Alerts is already on.")
                else
                    CAlerts_Config["extendedalerts"] = true
                    print("Extended Alerts has been turned on.")
                end
            end
        end
    elseif cmd[0] == "setfont" then
        if numargs ~= 3 or (cmd[1] ~= "title" and cmd[1] ~= "message") or (cmd[2] ~= "1" and cmd[2] ~= "2" and cmd[2] ~= "3") then
            print("Wrong Syntax. '/calerts setfont {title/message} {1/2/3}'")
        else
            CAlerts_Config["font" .. cmd[1]] = cmd[1] .. cmd[2] .. ".ttf"
            print("Successfully set " .. cmd[1] .. " font to #" .. cmd[2] .. ".\nNote: You must relog or type '/reloadui' for the font changes to take affect.")
        end
    elseif cmd[0] == "kbonly" then
        if numargs ~= 2 then
            print("Wrong Syntax. '/calerts kbonly {on/off}'")
        else
            if cmd[1] == "off" then
                if CAlerts_Config["kbonly"] == 0 then
                  print("Killing Blow Only mode is already off.")
                else -- Turn off
                  CAlerts_Config["kbonly"] = 0
                  print("Killing Blow Only mode has been turned off.")
                end
            else
                if CAlerts_Config["kbonly"] == 1 then
                  print("Killing Blow Only mode is already on.")
                else -- Turn on
                  CAlerts_Config["kbonly"] = 1
                  print("Killing Blow Only mode has been turned on.")
                end
            end
        end
    elseif cmd[0] == "help" then
        print(helpString)
    else
        print("Unknown command. Type '/calerts help' for a list of all commands.")
    end
end


-- @alertType string  The name of the alert type. (solokill, killingblow, damageassist, distractassist, healassist)
-- @isVictory bool  True for victory, false for death
-- @names array  Array of names to apply to alert message.
-- @iconFlags array  The icon flags to be passed to the alert.
-- @doPrint bool  False for show full alert, true for only print (i.e. in chat window only)
function CAlerts.processAlert(alertType, isVictory, names, iconFlags, doPrint)
  if CAlerts_Config["toggle"] == 0 then
    return
  end
  
  if CAlerts_Config["kbonly"] == 1 and alertType == "solokill" then
    alertType = "killingblow"
  end
  
  
  if CAlerts_Config["kbonly"] == 1 and (alertType ~= "killingblow" or not isVictory) then
    return
  end

  doPrint = doPrint or false
  iconFlags = iconFlags or {}
  
  local victoryType = "victory"
  if not isVictory then
    victoryType = "death"
  end
  
  local title = CAlerts_Config["strings"]["title"][victoryType][alertType]
  
  local message = CAlerts_Config["strings"]["message"][victoryType][alertType]
  for k,v in pairs(names) do
    message = message:gsub("{" .. k .. "}", v)
  end
  
  print("{" .. title .. "} " .. message)
  if not doPrint then
    theAlert:setMessage(title, message, iconFlags)
  end
end


-- @event initialisation after addon initialisation
table.insert(Event.Addon.Load.End, {
  function(addon) if addon == "CAlerts" then addonLoaded()
  end
  end, "CAlerts", "Initialise"
})
