--[[
	Project: NerfedUtils string library
	Version: 1.0.0
	Authors: NerfedWar (http://www.nerfedwar.net - nerfed.war@gmail.com)
	         Others (see individual function comments)

GPL3 is applicable unless a different license is stated within a function comment block.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

]]


-- Pads the left-side of a string with a specific set of characters.
string.lpad = function(str, len, char)
  if char == nil then char = ' ' end
  return string.rep(char, len - #tostring(str)) .. str
end

-- Pads the right-side of a string with a specific set of characters.
string.rpad = function(str, len, char)
  if char == nil then char = ' ' end
  return str .. string.rep(char, len - #tostring(str))
end


-- trim whitespace from both ends of string
string.trim = function(s)
  return s:find'^%s*$' and '' or s:match'^%s*(.*%S)'
end

-- trim whitespace from left end of string
string.triml = function(s)
  return s:match'^%s*(.*)'
end

-- trim whitespace from right end of string
string.trimr = function(s)
  return s:find'^%s*$' and '' or s:match'^(.*%S)'
end

-- This splits a string into a table of strings using a delimiter, which defaults to whitespace.
-- Author: sdonovan (http://snippets.luacode.org/?p=authors/sdonovan)
-- License: MIT/X11
string.split = function(s, re)
  local i1 = 1
  local ls = {}
  local append = table.insert
  if not re then re = '%s+' end
  if re == '' then return { s } end
  while true do
    local i2, i3 = s:find(re, i1)
    if not i2 then
      local last = s:sub(i1)
      if last ~= '' then append(ls, last) end
      if #ls == 1 and ls[1] == '' then
        return {}
      else
        return ls
      end
    end
    append(ls, s:sub(i1, i2 - 1))
    i1 = i3 + 1
  end
end

